from typing import Callable

WorkerIdGenerator = Callable[[], int]


def immutable_worker_id(worker_id: int) -> WorkerIdGenerator:
	"""
	create a worker id generator which always returns the given one
	"""
	return lambda: worker_id



from abc import abstractmethod
from datetime import datetime
from enum import Enum
from logging import getLogger
from os import getpid
from random import randrange
from socket import AF_INET, SOCK_DGRAM, socket
from threading import Thread
from typing import Callable, List, Optional

from time import sleep

from watchmen_model.common import Storable
from watchmen_utilities import get_current_time_in_seconds
from .snowflake_worker_id_generator import WorkerIdGenerator


class WorkerFirstDeclarationException(Exception):
	pass


class WorkerCreationException(Exception):
	pass


class WorkerDeclarationException(Exception):
	pass


def get_host_ip() -> str:
	s = None
	ip = None
	try:
		s = socket(AF_INET, SOCK_DGRAM)
		s.connect(('8.8.8.8', 80))
		ip = s.getsockname()[0]
	finally:
		if s is not None:
			s.close()
		return ip


class CompetitiveWorker(Storable):
	ip: Optional[str] = get_host_ip()
	processId: Optional[str] = str(getpid())
	dataCenterId: int = None
	workerId: int = None
	registeredAt: Optional[datetime] = get_current_time_in_seconds()
	lastBeatAt: datetime = None


def default_heart_beat_interval() -> int:
	"""
	:return: in seconds
	"""
	return 60


def default_worker_creation_retry_times() -> int:
	return 3


class CompetitiveWorkerShutdownSignal(Enum):
	EXIT = 1,
	EXCEPTION_RAISED = 2,


CompetitiveWorkerRestarter = Callable[[], None]
CompetitiveWorkerShutdownListener = Callable[
	[CompetitiveWorkerShutdownSignal, int, int, CompetitiveWorkerRestarter], None
]


class CompetitiveWorkerIdGenerator:
	worker: CompetitiveWorker = None
	firstDeclareTimes: int = 0

	def __init__(
			self,
			data_center_id: int = 0,
			heart_beat_interval: int = default_heart_beat_interval(),
			worker_creation_retry_times: int = default_worker_creation_retry_times(),
			shutdown_listener: CompetitiveWorkerShutdownListener = None
	):
		# will not check sanity of data center id here
		self.dataCenterId = data_center_id
		self.heartBeatInterval = heart_beat_interval
		self.workerCreationRetryTimes = worker_creation_retry_times
		self.handleShutdown = shutdown_listener
		self.try_create_worker()

	@abstractmethod
	def first_declare_myself(self, worker: CompetitiveWorker) -> None:
		"""
		first declare me, implement me
		"""
		pass

	def create_worker(self):
		# create a worker
		try:
			self.firstDeclareTimes += 1
			worker = CompetitiveWorker(dataCenterId=self.dataCenterId, workerId=self.create_worker_id())
			self.first_declare_myself(worker)
			return worker
		except WorkerFirstDeclarationException:
			if self.firstDeclareTimes <= self.workerCreationRetryTimes:
				return self.create_worker()
			else:
				raise WorkerCreationException(
					f'Failed to create worker[dataCenterId={self.dataCenterId}], '
					f'reaches maximum retry times[{self.workerCreationRetryTimes}]')

	def try_create_worker(self):
		self.firstDeclareTimes = 0
		self.worker = self.create_worker()
		del self.firstDeclareTimes
		# start heart beat
		Thread(target=CompetitiveWorkerIdGenerator.heart_beat, args=(self,), daemon=True).start()

	@staticmethod
	def random_worker_id() -> int:
		return randrange(0, 1024)

	@abstractmethod
	def acquire_alive_worker_ids(self) -> List[int]:
		"""
		acquire used worker ids, implement me
		:return: used worker ids
		"""
		pass

	def create_worker_id(self) -> int:
		alive_worker_ids = self.acquire_alive_worker_ids()
		# random a worker id, return it when it is not used
		new_worker_id = CompetitiveWorkerIdGenerator.random_worker_id()
		while new_worker_id in alive_worker_ids:
			new_worker_id = CompetitiveWorkerIdGenerator.random_worker_id()
		# return
		return new_worker_id

	@abstractmethod
	def declare_myself(self, worker: CompetitiveWorker) -> None:
		"""
		declare me is alive, implement me
		"""
		pass

	# noinspection PyUnreachableCode
	def heart_beat(self):
		try:
			while True:
				self.declare_myself(self.worker)
				sleep(self.heartBeatInterval)
		except Exception as e:
			getLogger(__name__).error(e, exc_info=True, stack_info=True)
			self.handleShutdown(
				CompetitiveWorkerShutdownSignal.EXCEPTION_RAISED,
				self.worker.dataCenterId, self.worker.workerId,
				self.try_create_worker
			)
		else:
			# heart beat stopped with no exception, release signal
			self.handleShutdown(
				CompetitiveWorkerShutdownSignal.EXIT,
				self.worker.dataCenterId, self.worker.workerId,
				self.try_create_worker
			)
		finally:
			# release in-memory worker, will raise exception only if somebody calls me later
			del self.worker
			getLogger(__name__).warning(f'Competitive worker id generator[{self.worker}] heart beat stopped.')

	def generate(self) -> int:
		"""
		generate snowflake worker id
		"""
		return self.worker.workerId


def competitive_worker_id(generator: CompetitiveWorkerIdGenerator) -> WorkerIdGenerator:
	"""
	create a worker id generator which delegate to given competitive generator
	"""
	return lambda: generator.generate()



from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, TypeVar, Union

from watchmen_model.common import DataModel, Pageable, Storable

"""
column name of storage entity, not for python object
"""
EntityColumnName = TypeVar('EntityColumnName', bound=str)
EntityColumnValue = Any
"""
entity name of storage entity, not for python object
"""
EntityName = TypeVar('EntityName', bound=str)
EntityRow = Dict[EntityColumnName, EntityColumnValue]
"""
entity can be a row or a base model
"""
Entity = Union[EntityRow, Storable]
"""
entity list can be list of rows or list of base models, cannot be mixed by rows and base models 
"""
EntityList = Union[List[EntityRow], List[Storable]]
EntityId = TypeVar('EntityId', str, int)


class EntityShaper:
	"""
	serializer/deserializer between python object and entity in storage
	"""

	@abstractmethod
	def serialize(self, entity: Entity) -> EntityRow:
		pass

	@abstractmethod
	def deserialize(self, row: EntityRow) -> Entity:
		pass


class EntityHelper(DataModel):
	name: EntityName
	shaper: EntityShaper


class EntityIdHelper(EntityHelper):
	idColumnName: EntityColumnName


class EntitySortMethod(str, Enum):
	ASC = 'asc',
	DESC = 'desc'


class EntitySortColumn(DataModel):
	name: EntityColumnName
	method: EntitySortMethod


EntitySort = List[EntitySortColumn]


class EntityCriteriaStatement(DataModel):
	pass


class EntityCriteriaOperator(str, Enum):
	IS_EMPTY = 'is-empty',
	IS_NOT_EMPTY = 'is-not-empty',
	IS_BLANK = 'is-blank',
	IS_NOT_BLANK = 'is-not-blank',
	EQUALS = 'equals',
	NOT_EQUALS = 'not-equals',
	LESS_THAN = 'less-than',
	LESS_THAN_OR_EQUALS = 'less-than-or-equals',
	GREATER_THAN = 'greater-than',
	GREATER_THAN_OR_EQUALS = 'greater-than-or-equals',
	IN = 'in',
	NOT_IN = 'not-in',
	LIKE = 'like',
	NOT_LIKE = 'not-like'


class ComputedLiteralOperator(str, Enum):
	ADD = 'add',
	SUBTRACT = 'subtract',
	MULTIPLY = 'multiply',
	DIVIDE = 'divide',
	MODULUS = 'modulus',
	YEAR_OF = 'year-of',
	HALF_YEAR_OF = 'half-year-of',
	QUARTER_OF = 'quarter-of',
	MONTH_OF = 'month-of',
	WEEK_OF_YEAR = 'week-of-year',
	WEEK_OF_MONTH = 'week-of-month',
	DAY_OF_MONTH = 'day-of-month',
	DAY_OF_WEEK = 'day-of-week',
	CASE_THEN = 'case-then',

	CONCAT = 'concat',
	YEAR_DIFF = 'year-diff',
	MONTH_DIFF = 'month-diff',
	DAY_DIFF = 'day-diff',
	FORMAT_DATE = 'format-date',

	CHAR_LENGTH = 'char-length',


class ComputedLiteral(DataModel):
	operator: ComputedLiteralOperator
	elements: List[Union[Literal, Tuple[EntityCriteriaStatement, Literal]]]


class ColumnNameLiteral(DataModel):
	entityName: Optional[EntityName]
	columnName: EntityColumnName


Literal = Union[EntityColumnValue, ComputedLiteral, ColumnNameLiteral]


class EntityCriteriaExpression(EntityCriteriaStatement):
	left: Literal
	operator: EntityCriteriaOperator = EntityCriteriaOperator.EQUALS
	right: Optional[Literal] = None


class EntityCriteriaJointConjunction(str, Enum):
	AND = 'and',
	OR = 'or'


class EntityCriteriaJoint(EntityCriteriaStatement):
	conjunction: EntityCriteriaJointConjunction = EntityCriteriaJointConjunction.AND
	children: List[EntityCriteriaStatement]


EntityCriteria = List[EntityCriteriaStatement]
EntityUpdate = Dict[EntityColumnName, EntityColumnValue]


class EntityDeleter(EntityHelper):
	criteria: Optional[EntityCriteria] = None


class EntityFinder(EntityHelper):
	criteria: Optional[EntityCriteria] = None
	sort: Optional[EntitySort] = None


class EntityDistinctValuesFinder(EntityFinder):
	distinctColumnNames: List[EntityColumnName] = None
	distinctValueOnSingleColumn: bool = False  # distinct value when it is True and only one column assigned


class EntityStraightColumn(DataModel):
	columnName: EntityColumnName  # original name
	alias: Optional[EntityColumnName]  # alias name


class EntityColumnAggregateArithmetic(str, Enum):
	COUNT = 'count',
	SUM = 'sum',
	AVG = 'avg',
	MAX = 'max',
	MIN = 'min'


class EntityStraightAggregateColumn(EntityStraightColumn):
	arithmetic: EntityColumnAggregateArithmetic


class EntityStraightValuesFinder(EntityFinder):
	straightColumns: List[EntityStraightColumn] = None


class EntityPager(EntityFinder):
	pageable: Pageable


class EntityUpdater(EntityHelper):
	criteria: Optional[EntityCriteria] = None
	update: EntityUpdate



from typing import Union

from watchmen_model.admin import Topic


def as_table_name(topic_or_name: Union[Topic, str]) -> str:
	if isinstance(topic_or_name, str):
		return f'topic_{topic_or_name.strip().lower()}'
	else:
		return f'topic_{topic_or_name.name.strip().lower()}'



from .competitive_worker_id_generator import competitive_worker_id, CompetitiveWorker, CompetitiveWorkerIdGenerator, \
	CompetitiveWorkerRestarter, CompetitiveWorkerShutdownListener, CompetitiveWorkerShutdownSignal, get_host_ip, \
	WorkerCreationException, WorkerDeclarationException, WorkerFirstDeclarationException
from .data_source_helper import DataSourceHelper
from .free_storage_types import FreeAggregateArithmetic, FreeAggregateColumn, FreeAggregatePager, FreeAggregator, \
	FreeColumn, FreeFinder, FreeJoin, FreeJoinType, FreePager
from .settings import ask_decimal_fraction_digits, ask_decimal_integral_digits
from .snowflake import InvalidSystemClockException, SnowflakeGenerator
from .snowflake_worker_id_generator import immutable_worker_id, WorkerIdGenerator
from .storage_based_worker_id_generator import COMPETITIVE_WORKER_SHAPER, CompetitiveWorkerShaper, \
	SNOWFLAKE_WORKER_ID_TABLE, StorageBasedWorkerIdGenerator
from .storage_exception import EntityNotFoundException, InsertConflictException, NoCriteriaForUpdateException, \
	NoFreeJoinException, OptimisticLockException, TooManyEntitiesFoundException, UnexpectedStorageException, \
	UnsupportedComputationException, UnsupportedCriteriaException, UnsupportedSortMethodException, \
	UnsupportedStraightColumnException
from .storage_spi import StorageSPI, TopicDataStorageSPI, TransactionalStorageSPI
from .storage_types import ColumnNameLiteral, ComputedLiteral, ComputedLiteralOperator, Entity, \
	EntityColumnAggregateArithmetic, EntityColumnName, EntityColumnValue, EntityCriteria, EntityCriteriaExpression, \
	EntityCriteriaJoint, EntityCriteriaJointConjunction, EntityCriteriaOperator, EntityCriteriaStatement, \
	EntityDeleter, EntityDistinctValuesFinder, EntityFinder, EntityHelper, EntityId, EntityIdHelper, EntityList, \
	EntityName, EntityPager, EntityRow, EntityShaper, EntitySort, EntitySortColumn, EntitySortMethod, \
	EntityStraightAggregateColumn, EntityStraightColumn, EntityStraightValuesFinder, EntityUpdate, EntityUpdater, \
	Literal
from .topic_utils import as_table_name



from time import time

from .snowflake_worker_id_generator import WorkerIdGenerator

# noinspection SpellCheckingInspection
TWEPOCH = 1420041600000
# 0 - 3, 4 data centers maximum
DATACENTER_ID_BITS = 2
# 0 - 1023, 1024 workers maximum per data center
WORKER_ID_BITS = 10
# 0 - 1023, 1024 sequence numbers per millisecond
SEQUENCE_BITS = 10
MAX_WORKER_ID = -1 ^ (-1 << WORKER_ID_BITS)  # 2**10-1 0b1111111111
MAX_DATACENTER_ID = -1 ^ (-1 << DATACENTER_ID_BITS)
MAX_SEQUENCE = -1 ^ (-1 << SEQUENCE_BITS)
WORKER_ID_SHIFT = SEQUENCE_BITS
DATACENTER_ID_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS
TIMESTAMP_LEFT_SHIFT = SEQUENCE_BITS + WORKER_ID_BITS + DATACENTER_ID_BITS


class InvalidSystemClockException(Exception):
	pass


def generate_timestamp() -> int:
	return int(time() * 1000)


def acquire_next_millisecond(last_timestamp):
	timestamp = generate_timestamp()

	while timestamp <= last_timestamp:
		timestamp = generate_timestamp()

	return timestamp


class SnowflakeGenerator:
	"""
	snowflake id generator
	"""

	def __init__(self, data_center_id: int = 0, generate_worker_id: WorkerIdGenerator = None):
		"""
		:param data_center_id: data center id, [0, 3]
		:param generate_worker_id: a function returns worker id, [0, 1023]
		"""

		# sanity check
		if data_center_id > MAX_DATACENTER_ID or data_center_id < 0:
			raise ValueError(
				f'Data center id invalid, it must be between [0, {MAX_DATACENTER_ID}] and passed by [{data_center_id}] .')

		# compute worker id
		assert generate_worker_id is not None, 'Worker id generator cannot be none.'
		worker_id = generate_worker_id()
		# sanity check
		if worker_id > MAX_WORKER_ID or worker_id < 0:
			raise ValueError(
				f'Worker id invalid, it must be between [0, {MAX_WORKER_ID}] and passed by [{worker_id}] .')

		# initialize
		self.dataCenterId = data_center_id
		self.workerId = worker_id
		self.sequence = 0
		self.lastTimestamp = -1

	def next_id(self) -> int:
		timestamp = generate_timestamp()

		if timestamp < self.lastTimestamp:
			# incorrect timestamp
			# raise InvalidSystemClockException
			# use current time stamp
			timestamp = self.lastTimestamp

		if timestamp == self.lastTimestamp:
			# exactly in same timestamp, increase sequence
			# and increase timestamp when sequence reaches the max value
			self.sequence = (self.sequence + 1) & MAX_SEQUENCE
			if self.sequence == 0:
				timestamp = acquire_next_millisecond(self.lastTimestamp)
				self.lastTimestamp = timestamp
		else:
			# already beyonds in-memory timestamp, reset in-memory
			self.sequence = 0
			self.lastTimestamp = timestamp

		return \
			((timestamp - TWEPOCH) << TIMESTAMP_LEFT_SHIFT) | \
			(self.dataCenterId << DATACENTER_ID_SHIFT) | \
			(self.workerId << WORKER_ID_SHIFT) | \
			self.sequence



from datetime import timedelta
from logging import getLogger
from typing import List

from watchmen_utilities import ArrayHelper, get_current_time_in_seconds
from .competitive_worker_id_generator import CompetitiveWorker, CompetitiveWorkerIdGenerator, \
	CompetitiveWorkerShutdownListener, default_heart_beat_interval, default_worker_creation_retry_times, \
	WorkerDeclarationException, WorkerFirstDeclarationException
from .storage_spi import TransactionalStorageSPI
from .storage_types import ColumnNameLiteral, EntityCriteriaExpression, EntityCriteriaOperator, \
	EntityDistinctValuesFinder, EntityFinder, EntityHelper, EntityRow, EntityShaper, EntityUpdater

SNOWFLAKE_WORKER_ID_TABLE = 'snowflake_competitive_workers'


class CompetitiveWorkerShaper(EntityShaper):
	def serialize(self, entity: CompetitiveWorker) -> EntityRow:
		return {
			'ip': entity.ip,
			'process_id': entity.processId,
			'data_center_id': entity.dataCenterId,
			'worker_id': entity.workerId,
			'registered_at': entity.registeredAt,
			'last_beat_at': entity.lastBeatAt
		}

	def deserialize(self, row: EntityRow) -> CompetitiveWorker:
		return CompetitiveWorker(
			ip=row.get('ip'),
			processId=row.get('process_id'),
			dataCenterId=row.get('data_center_id'),
			workerId=row.get('worker_id'),
			registeredAt=row.get('registered_at'),
			lastBeatAt=row.get('last_beat_at')
		)


COMPETITIVE_WORKER_SHAPER = CompetitiveWorkerShaper()


class StorageBasedWorkerIdGenerator(CompetitiveWorkerIdGenerator):
	"""
	Table(snowflake_worker_id)
		ip, process_id, data_center_id, worker_id, registered_at, last_beat_at
	"""

	def __init__(
			self,
			storage: TransactionalStorageSPI,
			data_center_id: int = 0,
			heart_beat_interval: int = default_heart_beat_interval(),
			worker_creation_retry_times: int = default_worker_creation_retry_times(),
			shutdown_listener: CompetitiveWorkerShutdownListener = None
	):
		self.storage = storage
		super().__init__(data_center_id, heart_beat_interval, worker_creation_retry_times, shutdown_listener)

	@staticmethod
	def is_abandoned(worker: CompetitiveWorker) -> bool:
		return (get_current_time_in_seconds() - worker.lastBeatAt).days >= 1

	def first_declare_myself(self, worker: CompetitiveWorker) -> None:
		self.storage.begin()

		try:
			existing_workers = self.storage.find(
				EntityFinder(
					name=SNOWFLAKE_WORKER_ID_TABLE,
					shaper=COMPETITIVE_WORKER_SHAPER,
					criteria=[
						EntityCriteriaExpression(
							left=ColumnNameLiteral(columnName='data_center_id'), right=worker.dataCenterId),
						EntityCriteriaExpression(left=ColumnNameLiteral(columnName='worker_id'), right=worker.workerId)
					]
				)
			)

			workers_count = len(existing_workers)
			if workers_count == 0:
				# worker not exists
				worker.lastBeatAt = get_current_time_in_seconds()
				# handle insert failed when other process already did it, may raise exception
				try:
					self.storage.insert_one(
						worker,
						EntityHelper(name=SNOWFLAKE_WORKER_ID_TABLE, shaper=COMPETITIVE_WORKER_SHAPER)
					)
				except Exception as e:
					getLogger(__name__).error(e, exc_info=True, stack_info=True)
					raise WorkerFirstDeclarationException(
						f'Failed to declare worker[dataCenterId={worker.dataCenterId}, workerId={worker.workerId}], '
						f'there might be an existing one in storage.')
			elif workers_count == 1:
				# noinspection PyTypeChecker
				existing_worker: CompetitiveWorker = existing_workers[0]
				if StorageBasedWorkerIdGenerator.is_abandoned(existing_worker):
					# worker last beat before 1 day, treat it as abandoned
					# replace it
					worker.lastBeatAt = get_current_time_in_seconds()
					updated_count = self.storage.update_only(
						EntityUpdater(
							name=SNOWFLAKE_WORKER_ID_TABLE,
							shaper=COMPETITIVE_WORKER_SHAPER,
							criteria=[
								EntityCriteriaExpression(
									left=ColumnNameLiteral(columnName='data_center_id'), right=self.dataCenterId),
								EntityCriteriaExpression(
									left=ColumnNameLiteral(columnName='worker_id'), right=worker.workerId),
								EntityCriteriaExpression(
									left=ColumnNameLiteral(columnName='last_beat_at'),
									operator=EntityCriteriaOperator.LESS_THAN_OR_EQUALS,
									right=(get_current_time_in_seconds() + timedelta(days=-1))
								)
							],
							update={
								'ip': worker.ip,
								'process_id': worker.processId,
								'registered_at': worker.registeredAt,
								'last_beat_at': worker.lastBeatAt
							}
						)
					)
					# handle update failed when other process already did it, may raise exception
					if updated_count == 0:
						# no worker had been updated, which means declaration is failed
						raise WorkerFirstDeclarationException(
							f'Failed to declare worker[dataCenterId={worker.dataCenterId}, workerId={worker.workerId}], '
							f'there might be an alive one or not exists in storage.')
				else:
					# the only worker is still alive
					raise WorkerFirstDeclarationException(
						f'Worker[dataCenterId={worker.dataCenterId}, workerId={worker.workerId}, lastBeatAt={worker.lastBeatAt}] '
						f'still alive.')
			else:
				# multiple workers found
				raise WorkerFirstDeclarationException(
					f'Multiple workers[dataCenterId={worker.dataCenterId}, workerId={worker.workerId}, count={workers_count}] '
					f'determined.')
			# commit data
			self.storage.commit_and_close()
		except Exception as e:
			# rollback data
			self.storage.rollback_and_close()
			# rethrow exception
			raise e

	def acquire_alive_worker_ids(self) -> List[int]:
		self.storage.begin()
		try:
			rows = self.storage.find_distinct_values(
				EntityDistinctValuesFinder(
					name=SNOWFLAKE_WORKER_ID_TABLE,
					shaper=COMPETITIVE_WORKER_SHAPER,
					# workers last beat at in 1 day, means still alive
					criteria=[
						EntityCriteriaExpression(
							left=ColumnNameLiteral(columnName='data_center_id'), right=self.dataCenterId),
						EntityCriteriaExpression(
							left=ColumnNameLiteral(columnName='last_beat_at'),
							operator=EntityCriteriaOperator.GREATER_THAN,
							right=(get_current_time_in_seconds() + timedelta(days=-1))
						)
					],
					distinctColumnNames=['worker_id']
				)
			)
			return ArrayHelper(rows).map(lambda x: x.workerId).to_list()
		finally:
			self.storage.close()

	def declare_myself(self, worker: CompetitiveWorker) -> None:
		self.storage.begin()
		try:
			updated_count = self.storage.update_only(
				EntityUpdater(
					name=SNOWFLAKE_WORKER_ID_TABLE,
					shaper=COMPETITIVE_WORKER_SHAPER,
					criteria=[
						EntityCriteriaExpression(
							left=ColumnNameLiteral(columnName='data_center_id'), right=self.dataCenterId),
						EntityCriteriaExpression(left=ColumnNameLiteral(columnName='worker_id'), right=worker.workerId)
					],
					update={'last_beat_at': get_current_time_in_seconds()}
				)
			)
			if updated_count == 0:
				raise WorkerDeclarationException(
					f'Failed to declare worker[dataCenterId={worker.dataCenterId}, workerId={worker.workerId}], '
					f'certain data not found in storage.')

			self.storage.commit_and_close()
		except Exception as e:
			self.storage.rollback_and_close()
			raise e



from abc import abstractmethod

from watchmen_model.system import DataSource
from .storage_spi import StorageSPI, TopicDataStorageSPI


class DataSourceHelper:
	def __init__(self, data_source: DataSource):
		self.dataSource = data_source

	@abstractmethod
	def acquire_storage(self) -> StorageSPI:
		pass

	@abstractmethod
	def acquire_topic_data_storage(self) -> TopicDataStorageSPI:
		pass



from enum import Enum
from typing import List, Optional

from watchmen_model.common import DataModel, Pageable
from .storage_types import ColumnNameLiteral, EntityCriteria, EntitySortColumn, Literal


class FreeAggregateArithmetic(str, Enum):
	NONE = 'none'
	COUNT = 'count'
	SUMMARY = 'sum'
	AVERAGE = 'avg'
	MAXIMUM = 'max'
	MINIMUM = 'min'


class FreeColumn(DataModel):
	literal: Literal
	alias: str
	arithmetic: Optional[FreeAggregateArithmetic] = None


class FreeJoinType(str, Enum):
	LEFT = 'left',
	RIGHT = 'right',
	INNER = 'inner',


class FreeJoin(DataModel):
	primary: ColumnNameLiteral
	secondary: Optional[ColumnNameLiteral] = None
	type: Optional[FreeJoinType] = None


class FreeFinder(DataModel):
	columns: List[FreeColumn] = None
	joins: List[FreeJoin] = None
	criteria: Optional[EntityCriteria] = None


class FreeAggregateColumn(DataModel):
	name: str  # name must match free column's index, such as column_1, column_2 (starts from 1)
	arithmetic: Optional[FreeAggregateArithmetic] = None
	alias: Optional[str] = None


class FreePager(FreeFinder):
	pageable: Pageable = None


class FreeAggregator(FreeFinder):
	highOrderAggregateColumns: List[FreeAggregateColumn]
	"""
	leave empty on entity name
	column name must match free column's index, such as column_1, column_2 (starts from 1)
	"""
	highOrderCriteria: Optional[EntityCriteria] = None
	"""
	column name must match free column's index, such as column_1, column_2 (starts from 1)
	"""
	highOrderSortColumns: List[EntitySortColumn]
	highOrderTruncation: Optional[int] = None


class FreeAggregatePager(FreeAggregator):
	pageable: Pageable = None



from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from watchmen_model.admin import Topic
from watchmen_model.common import DataPage
from .free_storage_types import FreeAggregatePager, FreeAggregator, FreeFinder, FreePager
from .storage_types import Entity, EntityDeleter, EntityDistinctValuesFinder, EntityFinder, EntityHelper, EntityId, \
	EntityIdHelper, EntityList, EntityPager, EntityStraightValuesFinder, EntityUpdater


class StorageSPI(ABC):
	@abstractmethod
	def connect(self) -> None:
		"""
		connect when not connected, or do nothing if connected.
		call close when want to connect again. connection is autocommit.
		this method can be called multiple times, if there is a connection existing, do nothing
		"""
		pass

	@abstractmethod
	def close(self) -> None:
		pass

	@abstractmethod
	def insert_one(self, one: Entity, helper: EntityHelper) -> None:
		pass

	@abstractmethod
	def insert_all(self, data: List[Entity], helper: EntityHelper) -> None:
		pass

	@abstractmethod
	def update_one(self, one: Entity, helper: EntityIdHelper) -> int:
		"""
		returns 0 when update none, or 1 when update one
		"""
		pass

	@abstractmethod
	def update_only(self, updater: EntityUpdater, peace_when_zero: bool = False) -> int:
		"""
		update only one, if update none or more than one item, raise exception
		"""
		pass

	@abstractmethod
	def update_only_and_pull(self, updater: EntityUpdater) -> Optional[Entity]:
		"""
		update only one, if update none or more than one item, raise exception
		return the one before update
		"""
		pass

	@abstractmethod
	def update(self, updater: EntityUpdater) -> int:
		pass

	@abstractmethod
	def update_and_pull(self, updater: EntityUpdater) -> EntityList:
		pass

	@abstractmethod
	def delete_by_id(self, entity_id: EntityId, helper: EntityIdHelper) -> int:
		"""
		returns 0 when delete none, or 1 when delete one
		"""
		pass

	@abstractmethod
	def delete_by_id_and_pull(self, entity_id: EntityId, helper: EntityIdHelper) -> Optional[Entity]:
		"""
		return deleted none when delete none, or deleted one when delete one
		"""
		pass

	@abstractmethod
	def delete_only(self, deleter: EntityDeleter) -> int:
		"""
		delete only one, if delete none or more than one item, raise exception
		"""
		pass

	@abstractmethod
	def delete_only_and_pull(self, deleter: EntityDeleter) -> Optional[Entity]:
		"""
		delete only one, if delete none or more than one item, raise exception
		return the one before delete
		"""
		pass

	@abstractmethod
	def delete(self, deleter: EntityDeleter) -> int:
		pass

	@abstractmethod
	def delete_and_pull(self, deleter: EntityDeleter) -> EntityList:
		pass

	@abstractmethod
	def find_by_id(self, entity_id: EntityId, helper: EntityIdHelper) -> Optional[Entity]:
		pass

	@abstractmethod
	def find_and_lock_by_id(self, entity_id: EntityId, helper: EntityIdHelper) -> Optional[Entity]:
		pass

	@abstractmethod
	def find_one(self, finder: EntityFinder) -> Optional[Entity]:
		pass

	@abstractmethod
	def find(self, finder: EntityFinder) -> EntityList:
		pass

	@abstractmethod
	def find_distinct_values(self, finder: EntityDistinctValuesFinder) -> EntityList:
		"""
		filled values with given distinct columns, returns an entity list.
		entity is deserialized by shaper
		"""
		pass

	@abstractmethod
	def find_straight_values(self, finder: EntityStraightValuesFinder) -> EntityList:
		"""
		fill values with given straight columns, returns an entity list
		entity will not be deserialized by shaper.
		And when there is aggregated columns, other columns will be used in group by
		"""
		pass

	@abstractmethod
	def find_all(self, helper: EntityHelper) -> EntityList:
		pass

	@abstractmethod
	def page(self, pager: EntityPager) -> DataPage:
		pass

	@abstractmethod
	def exists(self, finder: EntityFinder) -> bool:
		pass

	@abstractmethod
	def count(self, finder: EntityFinder) -> int:
		pass


class TransactionalStorageSPI(StorageSPI):
	"""
	Example:
		def a_func():
			# Declare a storage and begin it explicitly\n
			storage = TransactionStorageSPI()
			storage.begin()\n

			# Then do logic with transaction\n
			try:
				# do some logic\n
				storage.commit_and_close()
			except:
				storage.rollback_and_close()

			# Or do logic with no transaction.\n
			# Use this pattern only when no write logic included\n
			try:
				# do some read logic\n
			finally:
				storage.close()
	"""

	@abstractmethod
	def begin(self) -> None:
		pass

	@abstractmethod
	def commit_and_close(self) -> None:
		"""
		1. commit successfully -> close
		2. commit failed -> throw exception
		"""
		pass

	@abstractmethod
	def rollback_and_close(self) -> None:
		"""
		1. rollback successfully -> close
		2. rollback failed -> close
		"""
		pass


class TopicDataStorageSPI(TransactionalStorageSPI):
	@abstractmethod
	def register_topic(self, topic: Topic) -> None:
		pass

	@abstractmethod
	def create_topic_entity(self, topic: Topic) -> None:
		pass

	@abstractmethod
	def update_topic_entity(self, topic: Topic, original_topic: Topic) -> None:
		pass

	@abstractmethod
	def drop_topic_entity(self, topic_name: str) -> None:
		pass

	@abstractmethod
	def truncate(self, helper: EntityHelper) -> None:
		pass

	# noinspection PyMethodMayBeStatic
	def is_free_find_supported(self) -> bool:
		return True

	def append_topic_to_trino(self, topic: Topic) -> None:
		pass

	def drop_topic_from_trino(self, topic: Topic) -> None:
		pass

	@abstractmethod
	def free_find(self, finder: FreeFinder) -> List[Dict[str, Any]]:
		pass

	@abstractmethod
	def free_page(self, pager: FreePager) -> DataPage:
		pass

	@abstractmethod
	def free_aggregate_find(self, aggregator: FreeAggregator) -> List[Dict[str, Any]]:
		pass

	@abstractmethod
	def free_aggregate_page(self, pager: FreeAggregatePager) -> DataPage:
		pass



# noinspection DuplicatedCode
class InsertConflictException(Exception):
	pass


class OptimisticLockException(Exception):
	pass


class UnexpectedStorageException(Exception):
	pass


class UnsupportedCriteriaException(UnexpectedStorageException):
	pass


class UnsupportedComputationException(UnexpectedStorageException):
	pass


class UnsupportedStraightColumnException(UnexpectedStorageException):
	pass


class NoFreeJoinException(UnexpectedStorageException):
	pass


class NoCriteriaForUpdateException(UnexpectedStorageException):
	pass


class UnsupportedSortMethodException(UnexpectedStorageException):
	pass


class EntityNotFoundException(Exception):
	pass


class TooManyEntitiesFoundException(Exception):
	pass



from pydantic import BaseSettings


class StorageSettings(BaseSettings):
	DECIMAL_INTEGRAL_DIGITS: int = 24
	DECIMAL_FRACTION_DIGITS: int = 8


storage_settings = StorageSettings()


def ask_decimal_integral_digits() -> int:
	return storage_settings.DECIMAL_INTEGRAL_DIGITS


def ask_decimal_fraction_digits() -> int:
	return storage_settings.DECIMAL_FRACTION_DIGITS
